import logging
from asynctest import TestCase, Mock, patch
from barterdude.hooks import logging as hook_log
from barterdude.conf import BARTERDUDE_DEFAULT_APP_NAME


class TestLogging(TestCase):
    maxDiff = None

    def setUp(self):
        self.message = Mock()
        self.logging = hook_log.Logging()

    async def test_should_access_default_logger(self):
        log = hook_log.Logging("my_log", logging.DEBUG)
        logger = log.logger
        self.assertEqual(
            type(logger),
            logging.Logger
        )
        self.assertEqual(
            type(logger.handlers[0]),
            logging.StreamHandler
        )
        self.assertEqual(
            logger.name, f"{BARTERDUDE_DEFAULT_APP_NAME}.my_log")
        self.assertEqual(logger.level, logging.DEBUG)

    @patch("barterdude.hooks.logging.Logging.logger")
    @patch("barterdude.hooks.logging.repr")
    @patch("barterdude.hooks.logging.format_tb")
    async def test_should_log_on_connection_fail(
        self, format_tb, repr, logger
    ):
        retries = Mock()
        exception = Exception()
        await self.logging.on_connection_fail(exception, retries)
        repr.assert_called_once_with(exception)
        format_tb.assert_called_once_with(exception.__traceback__)
        logger.error.assert_called_once_with({
            "message": "Failed to connect to the broker",
            "retries": retries,
            "exception": repr.return_value,
            "traceback": format_tb.return_value,
        })


class TestLoggingNotRedacted(TestCase):
    maxDiff = None

    def setUp(self):
        hook_log.BARTERDUDE_LOG_REDACTED = False
        self.message = Mock()
        self.logging = hook_log.Logging()

    @patch("barterdude.hooks.logging.Logging.logger")
    @patch("barterdude.hooks.logging.json.dumps")
    async def test_should_log_before_consume(self, dumps, logger):
        await self.logging.before_consume(self.message)
        dumps.assert_called_once_with(self.message.body)
        logger.info.assert_called_once_with({
            "message": "Before consume message",
            "delivery_tag": self.message._delivery_tag,
            "message_body": dumps.return_value,
        })

    @patch("barterdude.hooks.logging.Logging.logger")
    @patch("barterdude.hooks.logging.json.dumps")
    async def test_should_log_on_success(self, dumps, logger):
        await self.logging.on_success(self.message)
        dumps.assert_called_once_with(self.message.body)
        logger.info.assert_called_once_with({
            "message": "Successfully consumed message",
            "delivery_tag": self.message._delivery_tag,
            "message_body": dumps.return_value,
        })

    @patch("barterdude.hooks.logging.Logging.logger")
    @patch("barterdude.hooks.logging.json.dumps")
    @patch("barterdude.hooks.logging.repr")
    @patch("barterdude.hooks.logging.format_tb")
    async def test_should_log_on_fail(self, format_tb, repr, dumps, logger):
        exception = Exception()
        await self.logging.on_fail(self.message, exception)
        dumps.assert_called_once_with(self.message.body)
        repr.assert_called_once_with(exception)
        format_tb.assert_called_once_with(exception.__traceback__)
        logger.error.assert_called_once_with({
            "message": "Failed to consume message",
            "delivery_tag": self.message._delivery_tag,
            "message_body": dumps.return_value,
            "exception": repr.return_value,
            "traceback": format_tb.return_value,
        })


class TestLoggingRedacted(TestCase):
    maxDiff = None

    def setUp(self):
        hook_log.BARTERDUDE_LOG_REDACTED = True
        self.message = Mock()
        self.logging = hook_log.Logging()

    @patch("barterdude.hooks.logging.Logging.logger")
    @patch("barterdude.hooks.logging.json.dumps")
    async def test_should_log_before_consume(self, dumps, logger):
        await self.logging.before_consume(self.message)
        dumps.assert_not_called()
        logger.info.assert_called_once_with({
            "message": "Before consume message",
            "delivery_tag": self.message._delivery_tag
        })

    @patch("barterdude.hooks.logging.Logging.logger")
    @patch("barterdude.hooks.logging.json.dumps")
    async def test_should_log_on_success(self, dumps, logger):
        await self.logging.on_success(self.message)
        dumps.assert_not_called()
        logger.info.assert_called_once_with({
            "message": "Successfully consumed message",
            "delivery_tag": self.message._delivery_tag
        })

    @patch("barterdude.hooks.logging.Logging.logger")
    @patch("barterdude.hooks.logging.json.dumps")
    @patch("barterdude.hooks.logging.repr")
    @patch("barterdude.hooks.logging.format_tb")
    async def test_should_log_on_fail(self, format_tb, repr, dumps, logger):
        exception = Exception()
        await self.logging.on_fail(self.message, exception)
        dumps.assert_not_called()
        repr.assert_called_once_with(exception)
        format_tb.assert_called_once_with(exception.__traceback__)
        logger.error.assert_called_once_with({
            "message": "Failed to consume message",
            "delivery_tag": self.message._delivery_tag,
            "exception": repr.return_value,
            "traceback": format_tb.return_value,
        })



from asynctest import TestCase, Mock
from freezegun import freeze_time
from barterdude.hooks.healthcheck import Healthcheck


@freeze_time()
class TestHealthcheck(TestCase):
    maxDiff = None

    def setUp(self):
        self.success_rate = 0.9
        self.health_window = 60.0
        self.healthcheck = Healthcheck(
            Mock(),
            "/healthcheck",
            self.success_rate,
            self.health_window
        )

    async def test_should_call_before_consume(self):
        await self.healthcheck.before_consume(None)

    async def test_should_pass_healthcheck_when_no_messages(self):
        response = await self.healthcheck(Mock())
        self.assertEqual(response.status, 200)
        self.assertEqual(response.content_type, "text/plain")
        self.assertEqual(
            response.body._value.decode('utf-8'),
            '{"message": "No messages in last 60.0s", "status": "ok"}'
        )

    async def test_should_pass_healthcheck_when_only_sucess(self):
        await self.healthcheck.on_success(None)
        response = await self.healthcheck(Mock())
        self.assertEqual(response.status, 200)
        self.assertEqual(response.content_type, "text/plain")
        self.assertEqual(
            response.body._value.decode('utf-8'),
            '{"message": "Success rate: 1.0 (expected: 0.9)", '
            '"fail": 0, "success": 1, "status": "ok"}'
        )

    async def test_should_pass_healthcheck_when_success_rate_is_high(self):
        await self.healthcheck.on_fail(None, None)
        for i in range(0, 9):
            await self.healthcheck.on_success(None)
        response = await self.healthcheck(Mock())
        self.assertEqual(response.status, 200)
        self.assertEqual(response.content_type, "text/plain")
        self.assertEqual(
            response.body._value.decode('utf-8'),
            '{"message": "Success rate: 0.9 (expected: 0.9)", '
            '"fail": 1, "success": 9, "status": "ok"}'
        )

    async def test_should_fail_healthcheck_when_only_fail(self):
        await self.healthcheck.on_fail(None, None)
        response = await self.healthcheck(Mock())
        self.assertEqual(response.status, 500)
        self.assertEqual(response.content_type, "text/plain")
        self.assertEqual(
            response.body._value.decode('utf-8'),
            '{"message": "Success rate: 0.0 (expected: 0.9)", '
            '"fail": 1, "success": 0, "status": "fail"}'
        )

    async def test_should_fail_healthcheck_when_success_rate_is_low(self):
        await self.healthcheck.on_success(None)
        await self.healthcheck.on_fail(None, None)
        response = await self.healthcheck(Mock())
        self.assertEqual(response.status, 500)
        self.assertEqual(response.content_type, "text/plain")
        self.assertEqual(
            response.body._value.decode('utf-8'),
            '{"message": "Success rate: 0.5 (expected: 0.9)", '
            '"fail": 1, "success": 1, "status": "fail"}'
        )

    async def test_should_fail_when_force_fail_is_called(self):
        self.healthcheck.force_fail()
        await self.healthcheck.on_success(None)
        response = await self.healthcheck(Mock())
        self.assertEqual(response.status, 500)
        self.assertEqual(response.content_type, "text/plain")
        self.assertEqual(
            response.body._value.decode('utf-8'),
            '{"message": "Healthcheck fail called manually", "status": "fail"}'
        )

    async def test_should_erase_old_messages(self):
        tick = 7
        with freeze_time(auto_tick_seconds=tick):
            for i in range(0, 10):
                await self.healthcheck.on_fail(None, None)
            await self.healthcheck.on_success(None)
            response = await self.healthcheck(Mock())
            self.assertEqual(
                response.body._value.decode('utf-8'),
                '{"message": "Success rate: 0.125 (expected: 0.9)", '
                '"fail": 7, "success": 1, "status": "fail"}'
            )

    async def test_should_fail_healthcheck_when_fail_to_connect(self):
        await self.healthcheck.on_connection_fail(None, 3)
        response = await self.healthcheck(Mock())
        self.assertEqual(
            response.body._value.decode('utf-8'),
            '{"message": "Reached max connection fails (3)", "status": "fail"}'
        )



from asynctest import TestCase, Mock, MagicMock

from barterdude.hooks.metrics.prometheus.definitions import Definitions
from prometheus_client.metrics import Counter, Histogram, REGISTRY


class TestDefinitions(TestCase):

    def setUp(self):
        self.metrics = MagicMock()
        self.metric = Mock()
        self.metrics.__getitem__.return_value = self.metric
        self.definitions = Definitions(
            registry=REGISTRY, metrics=self.metrics, labelkeys=["labelnames"]
        )

    def test_should_prepare_before_consume(self):
        name = "my_metric"
        self.definitions._prepare_before_consume(
            name, "namespace", "unit"
        )
        self.metrics.__setitem__.assert_called_once()
        mock_call = self.metrics.__setitem__.call_args[0]
        self.assertEqual(mock_call[0], "my_metric")
        self.assertEqual(type(mock_call[1]), Counter)

    def test_should_prepare_on_connection_fail(self):
        name = "my_metric"
        self.definitions._prepare_on_connection_fail(
            name, "namespace", "unit"
        )
        self.metrics.__setitem__.assert_called_once()
        mock_call = self.metrics.__setitem__.call_args[0]
        self.assertEqual(mock_call[0], "my_metric")
        self.assertEqual(type(mock_call[1]), Counter)

    def test_should_prepare_histogram_measure(self):
        name = "my_metric"
        self.definitions._prepare_histogram_measure(
            name, "namespace", "unit"
        )
        self.metrics.__setitem__.assert_called_once()
        mock_call = self.metrics.__setitem__.call_args[0]
        self.assertEqual(mock_call[0], "my_metric")
        self.assertEqual(type(mock_call[1]), Histogram)



from asynctest import TestCase, Mock
from barterdude.hooks import HttpHook


class TestHttpHook(TestCase):
    def setUp(self):
        self.app = Mock()

    def test_should_call_http_route_from_hook(self):
        hook = HttpHook(
            self.app,
            "/my_little_route"
        )
        self.app.add_endpoint.assert_called_once()
        self.app.add_endpoint.assert_called_with(
            routes=["/my_little_route"],
            methods=["GET"],
            hook=hook.__call__
        )

    async def test_should_fail_when_calling_unimplemented_methods(self):
        hook = HttpHook(
            self.app,
            "/my_little_route"
        )
        with self.assertRaises(NotImplementedError):
            await hook(Mock())
        with self.assertRaises(NotImplementedError):
            await hook.before_consume(None)
        with self.assertRaises(NotImplementedError):
            await hook.on_success(None)
        with self.assertRaises(NotImplementedError):
            await hook.on_fail(None, None)
        with self.assertRaises(NotImplementedError):
            await hook.on_connection_fail(None, None)



import freezegun
from asynctest import TestCase, Mock, patch

from barterdude.hooks.metrics.prometheus import Prometheus

from prometheus_client import (
    Counter,
    Gauge,
    Summary,
    Histogram,
    Info,
    Enum
)


class TestPrometheus(TestCase):
    def setUp(self):
        self.registry = Mock()
        self.labels = {"test": "my_test"}
        self.message = {"message": "test"}
        self.prometheus = Prometheus(
            labels=self.labels,
            registry=self.registry
        )

    @patch("barterdude.hooks.metrics.prometheus.Prometheus.metrics")
    async def test_should_call_prometheus_on_before_consume(
        self, mock_metric_obj
    ):
        mock_counter_histogram = Mock()
        mock_counter_histogram.observe = Mock()
        mock_labels = Mock()
        mock_labels.labels = Mock(
            return_value=mock_counter_histogram
        )
        mock_metric_obj.__getitem__ = Mock(return_value=mock_labels)
        await self.prometheus.before_consume(self.message)
        mock_labels.labels.assert_called_once()
        mock_labels.labels.assert_called_with(**self.labels)
        mock_counter_histogram.inc.assert_called_once()

    @patch("barterdude.hooks.metrics.prometheus.Prometheus.metrics")
    async def test_should_call_prometheus_on_success(self, mock_metric_obj):
        mock_counter_histogram = Mock()
        mock_counter_histogram.observe = Mock()
        mock_labels = Mock()
        mock_labels.labels = Mock(
            return_value=mock_counter_histogram
        )
        mock_metric_obj.__getitem__ = Mock(return_value=mock_labels)
        await self.prometheus.before_consume(self.message)
        await self.prometheus.on_success(self.message)
        self.assertEqual(mock_labels.labels.call_count, 2)
        mock_labels.labels.assert_called_with(
            error="", state="success", test="my_test"
        )
        self.assertEqual(mock_counter_histogram.inc.call_count, 1)
        mock_counter_histogram.observe.assert_called_once()

    @patch("barterdude.hooks.metrics.prometheus.Prometheus.metrics")
    async def test_should_call_prometheus_on_fail(self, mock_metric_obj):
        mock_counter_histogram = Mock()
        mock_counter_histogram.observe = Mock()
        mock_labels = Mock()
        mock_labels.labels = Mock(
            return_value=mock_counter_histogram
        )
        mock_metric_obj.__getitem__ = Mock(return_value=mock_labels)
        await self.prometheus.before_consume(self.message)
        await self.prometheus.on_fail(self.message, Exception())
        self.assertEqual(mock_labels.labels.call_count, 2)
        mock_labels.labels.assert_called_with(
            error="<class 'Exception'>", state="fail", test="my_test"
        )
        self.assertEqual(mock_counter_histogram.inc.call_count, 1)
        mock_counter_histogram.observe.assert_called_once()

    @patch("barterdude.hooks.metrics.prometheus.Prometheus.metrics")
    async def test_should_measure_time_on_success(self, mock_metric_obj):
        mock_histogram = Mock()
        mock_histogram.observe = Mock()
        mock_labels = Mock()
        mock_labels.labels = Mock(
            return_value=mock_histogram
        )
        mock_metric_obj.__getitem__ = Mock(return_value=mock_labels)
        with freezegun.freeze_time() as freeze:
            await self.prometheus.before_consume(self.message)
            freeze.tick()
            await self.prometheus.on_success(self.message)
        mock_histogram.observe.assert_called_once()
        mock_histogram.observe.assert_called_with(1.0)

    @patch("barterdude.hooks.metrics.prometheus.Prometheus.metrics")
    async def test_should_measure_time_on_fail(self, mock_metric_obj):
        mock_histogram = Mock()
        mock_histogram.observe = Mock()
        mock_labels = Mock()
        mock_labels.labels = Mock(
            return_value=mock_histogram
        )
        mock_metric_obj.__getitem__ = Mock(return_value=mock_labels)
        with freezegun.freeze_time() as freeze:
            await self.prometheus.before_consume(self.message)
            freeze.tick()
            await self.prometheus.on_fail(self.message, Exception())
        mock_histogram.observe.assert_called_once()
        mock_histogram.observe.assert_called_with(1.0)

    @patch("barterdude.hooks.metrics.prometheus.Metrics")
    async def test_should_remove_message_from_timer_on_complete(
        self, mock_metric_class
    ):
        await self.prometheus.before_consume(self.message)
        self.assertEqual(len(self.prometheus._msg_start), 1)
        await self.prometheus._on_complete(
            self.message, "success", Exception()
        )
        self.assertDictEqual(
            self.prometheus._msg_start,
            {}
        )

    @patch("barterdude.hooks.metrics.prometheus.Prometheus.metrics")
    async def test_should_increment_on_connection_fail_counter(self, metrics):
        counter = Mock()
        labels = Mock(labels=Mock(return_value=counter))
        metrics.__getitem__ = Mock(return_value=labels)
        await self.prometheus.on_connection_fail(Mock(), Mock())
        self.assertEqual(labels.labels.call_count, 1)
        labels.labels.assert_called_with(test='my_test')
        counter.inc.assert_called_once()

    def test_should_call_counter(self):
        self.assertTrue(
            isinstance(self.prometheus.metrics.counter(
                name="test", documentation="doc"
            ), Counter)
        )

    def test_should_call_gauge(self):
        self.assertTrue(
            isinstance(self.prometheus.metrics.gauge(
                name="test", documentation="doc"
            ), Gauge)
        )

    def test_should_call_summary(self):
        self.assertTrue(
            isinstance(self.prometheus.metrics.summary(
                name="test", documentation="doc"
            ), Summary)
        )

    def test_should_call_histogram(self):
        self.assertTrue(
            isinstance(self.prometheus.metrics.histogram(
                name="test", documentation="doc"
            ), Histogram)
        )

    def test_should_call_info(self):
        self.assertTrue(
            isinstance(self.prometheus.metrics.info(
                name="test", documentation="doc"
            ), Info)
        )

    def test_should_call_enum(self):
        self.assertTrue(
            isinstance(self.prometheus.metrics.enum(
                name="test", documentation="doc", states=Mock()
            ), Enum)
        )



from asynctest import TestCase, Mock
from prometheus_client.metrics import MetricWrapperBase, REGISTRY

from barterdude.hooks.metrics.prometheus.metrics import Metrics


class TestMetrics(TestCase):
    def test_should_set_and_get_items(self):
        metrics = Metrics(REGISTRY)
        metric_mock = Mock(spec=MetricWrapperBase)
        metrics["metric"] = metric_mock
        self.assertEqual(metrics["metric"], metric_mock)

    def test_should_raising_error_setting_duplicate_keys(self):
        metrics = Metrics(REGISTRY)
        metric_mock = Mock(spec=MetricWrapperBase)
        metrics["metric"] = metric_mock
        with self.assertRaises(ValueError):
            metrics["metric"] = metric_mock



import os
import asyncio

from asynctest import TestCase
from barterdude import BarterDude
from barterdude.monitor import Monitor
from barterdude.hooks.logging import Logging
from barterdude.message import ValidationException
from tests_unit.helpers import load_fixture
from tests_integration.helpers import ErrorHook
from asyncworker.connections import AMQPConnection
from random import choices
from string import ascii_uppercase


class RabbitMQConsumerTest(TestCase):
    use_default_loop = True

    async def setUp(self):
        self.input_queue = "test"
        self.output_exchange = "test_exchange"
        self.output_queue = "test_output"
        self.rabbitmq_host = os.environ.get("RABBITMQ_HOST", "127.0.0.1")
        self.barterdude_host = os.environ.get("BARTERDUDE_HOST", "127.0.0.1")

        self.connection = AMQPConnection(  # nosec
            name="barterdude-test",
            hostname=self.rabbitmq_host,
            username="guest",
            password="guest",
            prefetch=1
        )
        self.queue_manager = self.connection["/"]
        await self.queue_manager.connection._connect()
        await self.queue_manager.connection.channel.queue_declare(
            self.input_queue
        )
        await self.queue_manager.connection.channel.exchange_declare(
            self.output_exchange, "direct"
        )
        await self.queue_manager.connection.channel.queue_declare(
            self.output_queue
        )
        await self.queue_manager.connection.channel.queue_bind(
            self.output_queue, self.output_exchange, ""
        )

        self.messages = []
        for i in range(10):
            message = {"key": "".join(choices(ascii_uppercase, k=16))}
            self.messages.append(message)

        self.schema = load_fixture("schema.json")

        self.app = BarterDude(hostname=self.rabbitmq_host)

    async def tearDown(self):
        await self.app.shutdown()
        await self.queue_manager.connection.channel.queue_delete(
            self.input_queue
        )
        await self.queue_manager.connection.channel.queue_delete(
            self.output_queue
        )
        await self.queue_manager.connection.channel.exchange_delete(
            self.output_exchange
        )
        await self.queue_manager.connection.close()

    async def send_all_messages(self):
        futures = []
        for message in self.messages:
            futures.append(self.queue_manager.put(
                routing_key=self.input_queue,
                data=message
            ))
        await asyncio.gather(*futures)

    async def test_process_messages_successfully(self):
        received_messages = set()

        @self.app.consume_amqp([self.input_queue], coroutines=1)
        async def handler(message):
            nonlocal received_messages
            received_messages.add(message.body["key"])

        await self.app.startup()
        await self.send_all_messages()
        await asyncio.sleep(1)

        for message in self.messages:
            self.assertTrue(message["key"] in received_messages)

    async def test_process_messages_successfully_with_validation(self):
        received_messages = set()

        @self.app.consume_amqp(
            [self.input_queue], coroutines=1, validation_schema=self.schema)
        async def handler(message):
            nonlocal received_messages
            received_messages.add(message.body["key"])

        await self.app.startup()
        await self.send_all_messages()
        await asyncio.sleep(1)

        for message in self.messages:
            self.assertTrue(message["key"] in received_messages)

    async def test_process_messages_successfully_even_with_crashed_hook(self):
        received_messages = set()

        monitor = Monitor(ErrorHook())

        @self.app.consume_amqp(
            [self.input_queue], coroutines=1, monitor=monitor)
        async def handler(message):
            nonlocal received_messages
            received_messages.add(message.body["key"])

        await self.app.startup()
        await self.send_all_messages()
        await asyncio.sleep(1)

        for message in self.messages:
            self.assertTrue(message["key"] in received_messages)

    async def test_process_one_message_and_publish(self):
        @self.app.consume_amqp([self.input_queue], coroutines=1)
        async def forward(message):
            await self.app.publish_amqp(
                self.output_exchange,
                self.messages[1]
            )

        received_message = None

        @self.app.consume_amqp([self.output_queue], coroutines=1)
        async def handler(message):
            nonlocal received_message
            received_message = message.body

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)
        self.assertEqual(received_message, self.messages[1])

    async def test_process_message_requeue_with_requeue(self):
        handler_called = 0

        @self.app.consume_amqp([self.input_queue], coroutines=1)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise Exception()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)
        self.assertEqual(handler_called, 2)

    async def test_process_message_reject_with_requeue(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], requeue_on_fail=True,
            bulk_flush_interval=1, coroutines=1)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise Exception()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(2)
        self.assertEqual(handler_called, 2)

    async def test_process_message_reject_by_validation_with_requeue(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], requeue_on_validation_fail=True,
            bulk_flush_interval=1, coroutines=1)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise ValidationException()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(2)
        self.assertEqual(handler_called, 2)

    async def test_process_message_reject_without_requeue(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], requeue_on_fail=False,
            bulk_flush_interval=1, coroutines=1)
        async def handler(message):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise Exception()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(2)
        self.assertEqual(handler_called, 1)

    async def test_process_message_reject_by_validation_without_requeue(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], requeue_on_validation_fail=False)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise ValidationException()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(2)
        self.assertEqual(handler_called, 1)

    async def test_process_message_reject_by_validation_before_handler(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], validation_schema=self.schema)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data={"wrong": "key"}
        )
        await asyncio.sleep(1)
        self.assertEqual(handler_called, 0)

    async def test_process_messages_and_requeue_only_one(self):
        first_read = set()
        second_read = set()

        @self.app.consume_amqp(
            [self.input_queue],
            coroutines=len(self.messages),
            bulk_flush_interval=0.1
        )
        async def handler(message):
            nonlocal first_read
            nonlocal second_read
            value = message.body["key"]
            if value not in first_read:
                first_read.add(value)
                if message.body == self.messages[0]:
                    # process only messages[0] again
                    raise Exception()
            else:
                second_read.add(value)

        await self.app.startup()
        await self.send_all_messages()

        await asyncio.sleep(1)

        for message in self.messages:
            self.assertTrue(message["key"] in first_read)

        self.assertSetEqual(second_read, {self.messages[0]["key"]})

    async def test_fails_to_connect_to_rabbitmq(self):
        monitor = Monitor(Logging())

        self.app = BarterDude(hostname="invalid_host")

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            pass

        await self.app.startup()
        with self.assertLogs("barterdude") as cm:
            await asyncio.sleep(2)

        self.assertIn(
            "{'message': 'Failed to connect to the broker', 'retries': 1,",
            cm.output[0]
        )



import os
import aiohttp
import asyncio

from asynctest import TestCase
from barterdude import BarterDude
from barterdude.monitor import Monitor
from barterdude.hooks.healthcheck import Healthcheck
from barterdude.hooks import logging as hook_logging
from barterdude.hooks.metrics.prometheus import Prometheus
from tests_unit.helpers import load_fixture
from tests_integration.helpers import ErrorHook
from asyncworker.connections import AMQPConnection
from random import choices
from string import ascii_uppercase


class TestBarterDude(TestCase):
    use_default_loop = True

    async def setUp(self):
        self.input_queue = "test"
        self.output_exchange = "test_exchange"
        self.output_queue = "test_output"
        self.rabbitmq_host = os.environ.get("RABBITMQ_HOST", "127.0.0.1")
        self.barterdude_host = os.environ.get("BARTERDUDE_HOST", "127.0.0.1")

        self.connection = AMQPConnection(  # nosec
            name="barterdude-test",
            hostname=self.rabbitmq_host,
            username="guest",
            password="guest",
            prefetch=1
        )
        self.queue_manager = self.connection["/"]
        await self.queue_manager.connection._connect()
        await self.queue_manager.connection.channel.queue_declare(
            self.input_queue
        )
        await self.queue_manager.connection.channel.exchange_declare(
            self.output_exchange, "direct"
        )
        await self.queue_manager.connection.channel.queue_declare(
            self.output_queue
        )
        await self.queue_manager.connection.channel.queue_bind(
            self.output_queue, self.output_exchange, ""
        )

        self.messages = []
        for i in range(10):
            message = {"key": "".join(choices(ascii_uppercase, k=16))}
            self.messages.append(message)

        self.schema = load_fixture("schema.json")

        self.app = BarterDude(hostname=self.rabbitmq_host)

    async def tearDown(self):
        await self.app.shutdown()
        await self.queue_manager.connection.channel.queue_delete(
            self.input_queue
        )
        await self.queue_manager.connection.channel.queue_delete(
            self.output_queue
        )
        await self.queue_manager.connection.channel.exchange_delete(
            self.output_exchange
        )
        await self.queue_manager.connection.close()

    async def test_obtains_healthcheck(self):
        monitor = Monitor(Healthcheck(self.app))

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            pass

        await self.app.startup()

        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)

        async with aiohttp.ClientSession() as session:
            timeout = aiohttp.ClientTimeout(total=1)
            url = f'http://{self.barterdude_host}:8080/healthcheck'
            async with session.get(url, timeout=timeout) as response:
                status_code = response.status
                text = await response.text()

        self.assertEqual(status_code, 200)
        self.assertEqual(
            text,
            '{"message": "Success rate: 1.0 (expected: 0.95)", '
            '"fail": 0, "success": 1, "status": "ok"}'
        )

    async def test_obtains_healthcheck_even_with_crashed_hook(self):
        monitor = Monitor(ErrorHook(), Healthcheck(self.app))

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            pass

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)

        async with aiohttp.ClientSession() as session:
            timeout = aiohttp.ClientTimeout(total=1)
            url = f'http://{self.barterdude_host}:8080/healthcheck'
            async with session.get(url, timeout=timeout) as response:
                status_code = response.status
                text = await response.text()

        self.assertEqual(status_code, 200)
        self.assertEqual(
            text,
            '{"message": "Success rate: 1.0 (expected: 0.95)", '
            '"fail": 0, "success": 1, "status": "ok"}'
        )

    async def test_obtains_prometheus_metrics(self):
        labels = {"app_name": "barterdude_consumer"}
        monitor = Monitor(Prometheus(labels))

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            pass

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)

        async with aiohttp.ClientSession() as session:
            timeout = aiohttp.ClientTimeout(total=1)
            url = f'http://{self.barterdude_host}:8080/metrics'
            async with session.get(url, timeout=timeout) as response:
                status_code = response.status
                text = await response.text()

        self.assertEqual(status_code, 200)
        self.assertNotEqual(-1, text.find(
            'barterdude_received_number_before_consume_messages_total'
            '{app_name="barterdude_consumer"} 1.0'))
        self.assertNotEqual(-1, text.find(
            'barterdude_processing_message_seconds_bucket{app_name='
            '"barterdude_consumer",error="",le="0.025",state="success"} 1.0'
        ))
        self.assertNotEqual(-1, text.find(
            'barterdude_processing_message_seconds_count'
            '{app_name="barterdude_consumer",error="",state="success"} 1.0'
        ))
        self.assertNotEqual(-1, text.find(
            'asyncworker_http_requests_in_progress'
            '{method="GET",path="/metrics"} 1.0'
        ))

    async def test_print_logs_redacted(self):
        hook_logging.BARTERDUDE_LOG_REDACTED = True
        monitor = Monitor(hook_logging.Logging())
        error = Exception("raise expected")

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            if message.body == self.messages[0]:
                raise error

        with self.assertLogs("barterdude") as cm:
            await self.app.startup()
            await self.queue_manager.put(
                routing_key=self.input_queue,
                data=self.messages[0]
            )
            await asyncio.sleep(1)

        key = self.messages[0]["key"]
        error_str = repr(error)

        self.assertIn("'message': 'Before consume message'", cm.output[0])
        self.assertNotIn(
            f"'message_body': '{{\"key\": \"{key}\"}}'", cm.output[0]
        )
        self.assertIn("'delivery_tag': 1", cm.output[0])

        self.assertIn("'message': 'Failed to consume message'", cm.output[1])
        self.assertNotIn(
            f"'message_body': '{{\"key\": \"{key}\"}}'", cm.output[1]
        )
        self.assertIn("'delivery_tag': 1", cm.output[1])
        self.assertIn(f"'exception': \"{error_str}\"", cm.output[1])
        self.assertIn("'traceback': [", cm.output[1])

    async def test_print_logs(self):
        hook_logging.BARTERDUDE_LOG_REDACTED = False
        monitor = Monitor(hook_logging.Logging())
        error = Exception("raise expected")

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            if message.body == self.messages[0]:
                raise error

        with self.assertLogs("barterdude") as cm:
            await self.app.startup()
            await self.queue_manager.put(
                routing_key=self.input_queue,
                data=self.messages[0]
            )
            await asyncio.sleep(1)

        key = self.messages[0]["key"]
        error_str = repr(error)

        self.assertIn("'message': 'Before consume message'", cm.output[0])
        self.assertIn(
            f"'message_body': '{{\"key\": \"{key}\"}}'", cm.output[0]
        )
        self.assertIn("'delivery_tag': 1", cm.output[0])

        self.assertIn("'message': 'Failed to consume message'", cm.output[1])
        self.assertIn(
            f"'message_body': '{{\"key\": \"{key}\"}}'", cm.output[1]
        )
        self.assertIn("'delivery_tag': 1", cm.output[1])
        self.assertIn(f"'exception': \"{error_str}\"", cm.output[1])
        self.assertIn("'traceback': [", cm.output[1])
